/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.cocoon.pipeline.component.sax;

import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TemplatesHandler;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.cocoon.pipeline.util.StringRepresentation;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

public class XSLTTransformer extends AbstractTransformer {

    private final Log log = LogFactory.getLog(XSLTTransformer.class);
    private Map<String, Object> parameters;
    private URL source;

    public XSLTTransformer() {
        super();
    }

    public XSLTTransformer(URL source) {
        this(source, null);
    }

    public XSLTTransformer(URL source, Map<String, Object> parameters) {
        super();
        if (source == null) {
            throw new IllegalArgumentException("The parameter 'source' mustn't be null.");
        }

        this.parameters = parameters;
        this.source = source;
    }

    /**
     * Test if the name is a valid parameter name for XSLT
     */
    private static boolean isValidXSLTParameterName(String name) {
        return name.matches("[a-zA-Z_][\\w\\-\\.]*");
    }

    @Override
    public void setConfiguration(Map<String, ? extends Object> configuration) {
        this.source = (URL) configuration.get("source");

        this.parameters = new HashMap<String, Object>(configuration);
    }

    @Override
    protected void setXMLConsumer(XMLConsumer consumer) {
        TransformerHandler transformerHandler;
        try {
            transformerHandler = this.createTransformerHandler();
        } catch (Exception ex) {
            throw new RuntimeException("Could not initialize transformer handler.", ex);
        }

        final Map<String, Object> map = this.getLogicSheetParameters();
        if (map != null) {
            final Transformer transformer = transformerHandler.getTransformer();

            for (Entry<String, Object> entry : map.entrySet()) {
                transformer.setParameter(entry.getKey(), entry.getValue());
            }
        }

        final SAXResult result = new SAXResult();
        result.setHandler(consumer);
        // According to TrAX specs, all TransformerHandlers are LexicalHandlers
        result.setLexicalHandler(consumer);
        transformerHandler.setResult(result);

        TraxErrorListener traxErrorListener = new TraxErrorListener(this.log, this.source.toExternalForm());
        transformerHandler.getTransformer().setErrorListener(traxErrorListener);

        super.setXMLConsumer(new XMLConsumerAdapter(transformerHandler, transformerHandler));
    }

    private TransformerHandler createTransformerHandler() throws Exception {
        SAXTransformerFactory transformerFactory = (SAXTransformerFactory) TransformerFactory.newInstance();
        TemplatesHandler templatesHandler = transformerFactory.newTemplatesHandler();

        XMLReader xmlReader = XMLReaderFactory.createXMLReader();
        xmlReader.setContentHandler(templatesHandler);
        String systemId = this.source.toURI().toString();
        InputSource inputSource = new InputSource(systemId);
        xmlReader.parse(inputSource);

        // Create transformer handler
        final TransformerHandler handler = transformerFactory.newTransformerHandler(templatesHandler.getTemplates());
        handler.getTransformer().setErrorListener(
                new XSLTProcessorErrorListener(this.log, this.source.toExternalForm()));

        return handler;
    }

    private Map<String, Object> getLogicSheetParameters() {
        if (this.parameters == null) {
            return null;
        }

        Map<String, Object> result = new HashMap<String, Object>();

        for (Entry<String, Object> entry : this.parameters.entrySet()) {
            String name = entry.getKey();

            if (isValidXSLTParameterName(name)) {
                result.put(name, entry.getValue());
            }
        }

        return result;
    }

    @Override
    public String toString() {
        return StringRepresentation.buildString(this, "src=" + this.source);
    }
}
