/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.cocoon.rest.controller;

import java.util.HashMap;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.cocoon.rest.controller.method.Delete;
import org.apache.cocoon.rest.controller.method.Get;
import org.apache.cocoon.rest.controller.method.Head;
import org.apache.cocoon.rest.controller.method.Options;
import org.apache.cocoon.rest.controller.method.Post;
import org.apache.cocoon.rest.controller.method.Put;
import org.apache.cocoon.rest.controller.response.RestResponse;
import org.apache.cocoon.rest.controller.response.Status;

public class MethodDelegator {

    private Map<String, MethodDelegate> delegates = new HashMap<String, MethodDelegate>();

    public MethodDelegator() {
        super();

        this.delegates.put("DELETE", new DeleteDelegate());
        this.delegates.put("GET", new GetDelegate());
        this.delegates.put("HEAD", new HeadDelegate());
        this.delegates.put("OPTIONS", new OptionsDelegate());
        this.delegates.put("POST", new PostDelegate());
        this.delegates.put("PUT", new PutDelegate());
    }

    public RestResponse delegate(HttpServletRequest request, Object controller) throws Exception {
        if (request != null && request.getMethod() != null) {
            String method = this.getMethod(request);

            MethodDelegate methodDelegate = this.delegates.get(method);
            if (methodDelegate != null) {
                return methodDelegate.execute(controller);
            }
        }

        return new Status(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
    }

    private String getMethod(HttpServletRequest request) {
        String alternativeMethod = request.getParameter("_method");
        if (alternativeMethod != null) {
            if (this.delegates.keySet().contains(alternativeMethod.toUpperCase())) {
                return alternativeMethod.toUpperCase();
            }

            // TODO log the unsupported method name
        }

        return request.getMethod().toUpperCase();
    }

    private class DeleteDelegate extends MethodDelegate {

        @Override
        public RestResponse execute(Object controller) throws Exception {
            if (controller instanceof Delete) {
                Delete delete = (Delete) controller;
                return delete.doDelete();
            }
            return super.execute(controller);
        }
    }

    private class GetDelegate extends MethodDelegate {

        @Override
        public RestResponse execute(Object controller) throws Exception {
            if (controller instanceof Get) {
                Get get = (Get) controller;
                return get.doGet();
            }
            return super.execute(controller);
        }
    }

    private class HeadDelegate extends MethodDelegate {

        @Override
        public RestResponse execute(Object controller) throws Exception {
            if (controller instanceof Head) {
                Head head = (Head) controller;
                return head.doHead();
            }
            return super.execute(controller);
        }
    }

    private abstract class MethodDelegate {

        public RestResponse execute(Object controller) throws Exception {
            return new Status(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
        }
    }

    private class OptionsDelegate extends MethodDelegate {

        @Override
        public RestResponse execute(Object controller) throws Exception {
            if (controller instanceof Options) {
                Options options = (Options) controller;
                return options.doOptions();
            }
            return super.execute(controller);
        }
    }

    private class PostDelegate extends MethodDelegate {

        @Override
        public RestResponse execute(Object controller) throws Exception {
            if (controller instanceof Post) {
                Post post = (Post) controller;
                return post.doPost();
            }

            return super.execute(controller);
        }
    }

    private class PutDelegate extends MethodDelegate {

        @Override
        public RestResponse execute(Object controller) throws Exception {
            if (controller instanceof Put) {
                Put put = (Put) controller;
                return put.doPut();
            }
            return super.execute(controller);
        }
    }
}
